@@ -1,7 +1,9 @@
 package Ace;
 
 use strict;
-use Carp 'croak';
+use Carp qw(croak carp cluck);
+use WeakRef;
+
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $Error);
 
 require Exporter;
@@ -17,7 +19,7 @@ use overload
 
 # Optional exports
 @EXPORT_OK = qw(rearrange ACE_PARSE);
-$VERSION = '1.86';
+$VERSION = '1.89';
 
 use constant STATUS_WAITING => 0;
 use constant STATUS_PENDING => 1;
@@ -30,6 +32,20 @@ use constant DEFAULT_SOCKET => 2005;    # socket server
 require Ace::Iterator;
 eval qq{use Ace::Freesubs};  # XS file, may not be available
 
+# Map database names to objects (to fix file-caching issue)
+my %NAME2DB;
+
+# internal cache of objects
+my %MEMORY_CACHE;
+
+# Debugging level
+my $DEBUG_LEVEL;
+
+my %DEFAULT_CACHE_PARAMETERS = (
+				default_expires_in  => '1 day',
+				auto_purge_interval => '12 hours',
+				);
+
 # Preloaded methods go here.
 $Error = '';
 
@@ -45,7 +61,7 @@ sub connect {
   my $class = shift;
   my ($host,$port,$user,$pass,$path,$program,
       $objclass,$timeout,$query_timeout,$database,
-      $server_type,$url,$u,$p,$other);
+      $server_type,$url,$u,$p,$cache,$other);
 
   # one-argument single "URL" form
   if (@_ == 1) {
@@ -54,12 +70,12 @@ sub connect {
 
   # multi-argument (traditional) form
   ($host,$port,$user,$pass,
-   $path,$objclass,$timeout,$query_timeout,$url,$other) = 
+   $path,$objclass,$timeout,$query_timeout,$url,$cache,$other) = 
      rearrange(['HOST','PORT','USER','PASS',
 		'PATH','CLASS','TIMEOUT',
-		'QUERY_TIMEOUT','URL'],@_);
+		'QUERY_TIMEOUT','URL','CACHE'],@_);
 
-  ($host,$port,$u,$p,$server_type) = $class->process_url($url) 
+  ($host,$port,$u,$pass,$p,$server_type) = $class->process_url($url) 
     or croak "Usage:  Ace->connect(-host=>\$host,-port=>\$port [,-path=>\$path]\n"
       if defined $url;
 
@@ -88,20 +104,25 @@ sub connect {
     return;
   }
 
-  my $self = bless {
-		    'database'=> $database,
-		    'host'   => $host,
-		    'port'   => $port,
-		    'path'   => $path,
-		    'class'  => $objclass || 'Ace::Object',
-		    'timeout' => $query_timeout,
-		    'user'    => $user,
-		    'pass'    => $pass,
-		    'other'  => $other,
-		    'date_style' => 'java',
-		    'auto_save' => 0,
-		   },ref($class)||$class;
+  my $contents = {
+		  'database'=> $database,
+		  'host'   => $host,
+		  'port'   => $port,
+		  'path'   => $path,
+		  'class'  => $objclass || 'Ace::Object',
+		  'timeout' => $query_timeout,
+		  'user'    => $user,
+		  'pass'    => $pass,
+		  'other'  => $other,
+		  'date_style' => 'java',
+		  'auto_save' => 0,
+		 };
+
+  my $self = bless $contents,ref($class)||$class;
   eval "require $self->{class}" or warn $@;
+
+  $self->_create_cache($cache) if $cache;
+  $self->name2db("$self",$self);
   return $self;
 }
 
@@ -114,7 +135,7 @@ sub reopen {
     $database = $class->connect(-path=>$self->{path},%{$self->other});
   } else {
     $database = $class->connect($self->{host},$self->{port}, $self->{timeout},
-				      $self->{user},$self->{pass},%{$self->{other}});
+				$self->{user},$self->{pass},%{$self->{other}});
   }
   unless ($database) {
     $Ace::Error = "Couldn't open database";
@@ -137,14 +158,14 @@ sub class {
 sub process_url {
   my $class = shift;
   my $url = shift;
-  my ($host,$port,$user,$path,$server_type) = ('','','','','');
+  my ($host,$port,$user,$pass,$path,$server_type) = ('','','','','','');
 
   if ($url) {  # look for host:port
     local $_ = $url;
     if (m!^rpcace://([^:]+):(\d+)$!) {  # rpcace://localhost:200005
       ($host,$port) = ($1,$2);
       $server_type = 'Ace::RPC';
-    } elsif (m!^sace://(\w+)\@([^:]+):(\d+)$!) { # sace://user@localhost:2005
+    } elsif (m!^sace://([\w:]+)\@([^:]+):(\d+)$!) { # sace://user@localhost:2005
       ($user,$host,$port) = ($1,$2,$3);
       $server_type = 'Ace::SocketServer';
     } elsif (m!^sace://([^:]+):(\d+)$!) { # sace://localhost:2005
@@ -161,7 +182,11 @@ sub process_url {
     }
   }
 
-  return ($host,$port,$user,$path,$server_type);  
+  if ($user =~ /:/) {
+    ($user,$pass) = split /:/,$user;
+  }
+
+  return ($host,$port,$user,$pass,$path,$server_type);  
 
 }
 
@@ -176,17 +201,151 @@ sub db {
 sub model {
   my $self = shift;
   require Ace::Model;
-  my $model = shift;
-  return $self->{'models'}{$model} ||= 
-    Ace::Model->new($self->raw_query("model \"$model\""));
+  my $model       = shift;
+  my $break_cycle = shift;  # for breaking cycles when following #includes
+  my $key = join(':',$self,'MODEL',$model);
+  $self->{'models'}{$model} ||= eval{$self->cache->get($key)};
+  unless ($self->{models}{$model}) {
+    $self->{models}{$model} =
+      Ace::Model->new($self->raw_query("model \"$model\""),$self,$break_cycle);
+    eval {$self->cache->set($key=>$self->{models}{$model})};
+  }
+  return $self->{'models'}{$model};
 }
-	   
+
+# cached get
+# pass "1" for fill to get a full fill
+# pass any other true value to get a tag fill
+sub get {
+  my $self = shift;
+  my ($class,$name,$fill) = @_;
+
+  # look in caches first
+  my $obj = $self->memory_cache_fetch($class=>$name) 
+    || $self->file_cache_fetch($class=>$name);
+  return $obj if $obj;
+
+  # _acedb_get() does the caching
+  $obj = $self->_acedb_get($class,$name,$fill) or return;
+  $obj;
+}
+
+sub _acedb_get {
+  my $self = shift;
+  my ($class,$name,$filltag) = @_;
+  return unless $self->count($class,$name) >= 1;
+
+  #return $self->{class}->new($class,$name,$self,1) unless $filltag;
+  return ($self->_list)[0] unless $filltag;
+
+  if (defined $filltag && $filltag eq '1') {  # full fill
+    return $self->_fetch();
+  } else {
+    return $self->_fetch(undef,undef,$filltag);
+  }
+}
+
+
+#### CACHE AND CARRY CODE ####
+# Be very careful here.  The key used for the memory cache is in the format
+# db:class:name, but the key used for the file cache is in the format class:name.
+# The difference is that the filecache has a built-in namespace but the memory
+# cache doesn't.
+sub memory_cache_fetch {
+  my $self = shift;
+  my ($class,$name) = @_;
+  my $key = join ":",$self,$class,$name;
+  return unless defined $MEMORY_CACHE{$key};
+  carp "memory_cache hit on $key = ",overload::StrVal($MEMORY_CACHE{$key})
+    if Ace->debug;
+  return $MEMORY_CACHE{$key};
+}
+
+sub memory_cache_store {
+  my $self = shift;
+  croak "Usage: memory_cache_store(\$obj)" unless @_ == 1;
+  my $obj = shift;
+  my $key = join ':',$obj->db,$obj->class,$obj->name;
+  carp "memory_cache store on $key = ",overload::StrVal($obj) if Ace->debug;
+  weaken($MEMORY_CACHE{$key} = $obj);
+}
+
+sub memory_cache_clear {
+    my $self = shift;
+    %MEMORY_CACHE = ();
+}
+
+sub memory_cache_delete {
+  my $package = shift;
+  my $obj = shift or croak "Usage: memory_cache_delete(\$obj)";
+  my $key = join ':',$obj->db,$obj->class,$obj->name;
+  delete $MEMORY_CACHE{$key};
+}
+
+# Call as:
+# $ace->file_cache_fetch($class=>$id)
+sub file_cache_fetch {
+  my $self = shift;
+  my ($class,$name) = @_;
+  my $key = join ':',$class,$name;
+  my $cache = $self->cache or return;
+  my $obj   = $cache->get($key);
+#   warn "cache ",$obj?'hit':'miss'," on '$key'\n";
+  if ($obj && !exists $obj->{'.root'}) {  # consistency checks
+    require Data::Dumper;
+    warn "CACHE BUG! Discarding inconsistent object $obj\n";
+    warn Data::Dumper->Dump([$obj],['obj']);
+    $cache->remove($key);
+    return;
+  }
+  #carp "cache ",$obj?'hit':'miss'," on '$key'\n" if Ace->debug;
+  $self->memory_cache_store($obj) if $obj;
+  $obj;
+}
+
+# call as
+# $ace->file_cache_store($obj);
+sub file_cache_store {
+  my $self = shift;
+  my $obj  = shift;
+  my $key = join ':',$obj->class,$obj->name;
+  my $cache = $self->cache or return;
+
+#  carp "caching $key obj=",overload::StrVal($obj),"\n" if Ace->debug;
+#   warn "caching $key obj=",overload::StrVal($obj),"\n";
+  if ($key eq ':') {  # something badly wrong
+    cluck "NULL OBJECT";
+  }
+  $cache->set($key,$obj);
+}
+
+sub file_cache_delete {
+  my $self = shift;
+  my $obj = shift;
+  my $key = join ':',$obj->class,$obj->name;
+  my $cache = $self->cache or return;
+
+  carp "deleting $key obj=",overload::StrVal($obj),"\n" if Ace->debug;
+  $cache->remove($key,$obj);
+}
+
+#### END: CACHE AND CARRY CODE ####
+
+
 # Fetch one or a group of objects from the database
 sub fetch {
   my $self = shift;
   my ($class,$pattern,$count,$offset,$query,$filled,$total,$filltag) =  
     rearrange(['CLASS',['NAME','PATTERN'],'COUNT','OFFSET','QUERY',
 	       ['FILL','FILLED'],'TOTAL','FILLTAG'],@_);
+
+  if (defined $class
+      && defined $pattern
+      && $pattern !~ /[\?\*]/
+      && !wantarray) {
+    return $self->get($class,$pattern);
+  }
+
   $offset += 0;
   $pattern ||= '*';
   $pattern = Ace->freeprotect($pattern);
@@ -197,6 +356,8 @@ sub fetch {
   } else {
     croak "must call fetch() with the -class or -query arguments";
   }
+
+
   my $r = $self->raw_query($query);
 
   my ($cnt) = $r =~ /Found (\d+) objects/m;
@@ -212,9 +373,44 @@ sub fetch {
   } else {
     @h = $filled ? $self->_fetch($count,$offset) : $self->_list($count,$offset);
   }
+
   return wantarray ? @h : $h[0];
 }
 
+sub cache    { 
+  my $self = shift;
+  my $d    = $self->{filecache};
+  $self->{filecache} = shift if @_;
+  $d;
+}
+
+sub _create_cache {
+  my $self   = shift;
+  my $params = shift;
+  $params    = {} if $params and !ref $params;
+
+  return unless eval {require Cache::SizeAwareFileCache};  # not installed
+
+  (my $namespace = "$self") =~ s!/!_!g;
+  my %cache_params = (
+		      namespace    => $namespace,
+		      %DEFAULT_CACHE_PARAMETERS,
+		      %$params,
+		     );
+  my $cache_obj = Cache::SizeAwareFileCache->new(\%cache_params);
+  $self->cache($cache_obj);
+}
+
+# class method
+sub name2db {
+  shift;
+  my $name = shift;
+  return unless defined $name;
+  my $d = $NAME2DB{$name};
+  weaken($NAME2DB{$name} = shift) if @_;
+  $d;
+}
+
 # make a new object using indicated class and name pattern
 sub new {
   my $self = shift;
@@ -268,6 +464,7 @@ sub keyset {
   return $self->_list;
 }
 
+
 #########################################################
 # These functions are for low-level (non OO) access only.
 # This is for low-level access only.
@@ -275,7 +472,7 @@ sub show {
     my ($self,$class,$pattern,$tag) = @_;
     $Ace::Error = '';
     return unless $self->count($class,$pattern);
-    
+
     # if we get here, then we've got some data to return.
     my @result;
     my $ts = $self->{'timestamps'} ? '-T' : '';
@@ -303,6 +500,7 @@ sub read_object {
 # do a query, and return the result immediately
 sub raw_query {
   my ($self,$query,$no_alert,$parse) = @_;
+  warn "raw_query($query)\n" if $self->debug;
   $self->_alert_iterators unless $no_alert;
   $self->{database}->query($query, $parse ? ACE_PARSE : () );
   return $self->read_object;
@@ -328,6 +526,8 @@ sub close {
 
 sub DESTROY { 
   my $self = shift;
+  return if caller() =~ /^Cache\:\:/;
+  warn "$self->DESTROY at ", join ' ',caller() if Ace->debug;
   $self->close;
 }
 
@@ -395,7 +595,14 @@ sub _list {
   foreach (split("\n",$result)) {
     my ($class,$name) = Ace->split($_);
     next unless $class and $name;
-    push(@result,$self->{'class'}->new($class,$name,$self,1));
+    my $obj = $self->memory_cache_fetch($class,$name);
+    $obj  ||= $self->file_cache_fetch($class,$name);
+    unless ($obj) {
+      $obj = $self->{'class'}->new($class,$name,$self,1);
+      $self->memory_cache_store($obj);
+      $self->file_cache_store($obj);
+    }
+    push @result,$obj;
   }
   return @result;
 }
@@ -422,6 +629,15 @@ sub _fetch {
       $tree = $obj;
     }
   }
+  # now recache 'em
+  for (@result) {
+    if (my $obj = $self->memory_cache_store($_)) {
+      %$obj = %$_ unless $obj->filled;  # contents copy -- replace partial object with full object
+      $_ = $obj;
+    } else {
+      $self->memory_cache_store($_);
+    }
+  }
   return wantarray ? @result : $result[0];
 }
 
@@ -443,7 +659,7 @@ sub _fetch_chunk {
 sub _alert_iterators {
   my $self = shift;
   foreach (keys %{$self->{iterators}}) {
-    $self->{iterators}{$_}->invalidate;
+    $self->{iterators}{$_}->invalidate if $self->{iterators}{$_};
   }
   undef $self->{active_list};
 }
@@ -451,7 +667,7 @@ sub _alert_iterators {
 sub asString {
   my $self = shift;
   return "tace://$self->{path}" if $self->{'path'};
-  my $server = $self->db->isa('Ace::SocketServer') ? 'sace' : 'rpcace';
+  my $server = $self->db && $self->db->isa('Ace::SocketServer') ? 'sace' : 'rpcace';
   return "$server://$self->{host}:$self->{port}" if $self->{'host'};
   return ref $self;
 }
@@ -468,7 +684,43 @@ sub cmp {
 }
 
 
+# Count the objects matching pattern without fetching them.
+sub count {
+  my $self = shift;
+  my ($class,$pattern,$query) = rearrange(['CLASS',
+					   ['NAME','PATTERN'],
+					   'QUERY'],@_);
+  $Ace::Error = '';
 
+  # A special case occurs when we have already fetched this
+  # object and it is already on the active list.  In this
+  # case, we do not need to recount.
+  $query   = '' unless defined $query;
+  $pattern = '' unless defined $pattern;
+  $class   = '' unless defined $class;
+
+  my $active_tag = "$class$pattern$query";
+  if (defined $self->{'active_list'} &&
+      defined ($self->{'active_list'}->{$active_tag})) {
+    return $self->{'active_list'}->{$active_tag};
+  }
+
+  if ($query) {
+    $query = "query $query" unless $query=~/^query\s/;
+  } else {
+    $pattern =~ tr/\n//d;
+    $pattern ||= '*';
+    $pattern = Ace->freeprotect($pattern);
+    $query = "find $class $pattern";
+  }
+  my $result = $self->raw_query($query);
+#  unless ($result =~ /Found (\d+) objects/m) {
+  unless ($result =~ /(\d+) Active Objects/m) {
+    $Ace::Error = 'Unexpected close during find';
+    return;
+  }
+  return $self->{'active_list'}->{$active_tag} = $1;
+}
 
 1;
 
@@ -555,7 +807,7 @@ classes.
     # remote database
     $db = Ace->connect(-host  =>  'beta.crbm.cnrs-mop.fr',
                        -port  =>  20000100);
-    
+
     # local (non-server) database
     $db = Ace->connect(-path  =>  '/usr/local/acedb);
 
@@ -572,7 +824,9 @@ full syntax is as follows:
 		       -program =>$local_connection_program
                        -class =>  $object_class,
 		       -timeout => $timeout,
-		       -query_timeout => $query_timeout);
+		       -query_timeout => $query_timeout
+		       -cache        => {cache parameters},
+		      );
 
 The connect() method uses a named argument calling style, and
 recognizes the following arguments:
@@ -601,6 +855,45 @@ provided, will attempt an anonymous login.
 
 Password to log in with (when using socket server).
 
+=item B<-url>
+
+An Acedb URL that combines the server type, host, port, user and
+password in a single string.  See the connect() method's "single
+argument form" description.
+
+=item B<-cache>
+
+AcePerl can use the Cache::SizeAwareFileCache module to cache objects
+to disk. This can result in dramatically increased performance in
+environments such as web servers in which the same Acedb objects are
+frequently reused.  To activate this mechanism, the
+Cache::SizeAwareFileCache module must be installed, and you must pass
+the -cache argument during the connect() call.
+
+The value of -cache is a hash reference containing the arguments to be
+passed to Cache::SizeAwareFileCache.  For example:
+
+   -cache => {
+              cache_root         => '/usr/tmp/acedb',
+              cache_depth        => 4,
+              default_expires_in => '1 hour'
+              }
+
+If not otherwise specified, the following cache parameters are assumed:
+
+       Parameter               Default Value
+       ---------               -------------
+       namespace               Server URL (e.g. sace://localhost:2005)
+       cache_root              /tmp/FileCache (dependent on system temp directory)
+       default_expires_in      1 day
+       auto_purge_interval     12 hours
+
+By default, the cache is not size limited (the "max_size" property is
+set to $NO_MAX_SIZE).  To adjust the size you may consider calling the
+Ace object's cache() method to retrieve the physical cache and then
+calling the cache object's limit_size($max_size) method from time to
+time.  See L<Cache::SizeAwareFileCache> for more details.
+
 =item B<-program>
 
 By default AcePerl will use its internal compiled code calls to
@@ -856,7 +1149,38 @@ Example:
 
 If your request is likely to retrieve very many objects, fetch() many
 consume a lot of memory, even if B<-fill> is false.  Consider using
-B<fetch_many()> instead (see below).
+B<fetch_many()> instead (see below).  Also see the get() method, which
+is equivalent to the simple two-argument form of fetch().
+
+=item get() method
+
+   $object = $db->get($class,$name [,$fill]);
+
+The get() method will return one and only one AceDB object
+identified by its class and name.  The optional $fill argument can be
+used to control how much data is retrieved from the database. If $fill
+is absent or undefined, then the method will return a lightweight
+"stub" object that is filled with information as requested in a lazy
+fashion. If $fill is the number "1" then the retrieved object contains
+all the relevant information contained within the database.  Any other
+true value of $fill will be treated as a tag name: the returned object
+will be prefilled with the subtree to the right of that tag.
+
+Examples:
+
+   # return lightweight stub for Author object "Sulston JE."
+   $author = $db->get(Author=>'Sulston JE');
+
+   # return heavyweight object
+   $author = $db->get(Author=>'Sulston JE',1);
+
+   # return object containing the Address subtree
+   $author = $db->get(Author=>'Sulston JE','Address');
+
+The get() method is equivalent to this form of the fetch()
+method:
+
+   $object = $db->fetch($class=>$name);
 
 =head2 aql() method
 
@@ -949,7 +1273,7 @@ Any parse error messages are accumulated in Ace->error().
 
 =head2 new() method
 
-  $object = $db->parse($class => $name);
+  $object = $db->new($class => $name);
 
 This method creates a new object in the database of type $class and
 name $name.  If successful, it returns the newly-created object.
@@ -1073,6 +1397,7 @@ objects.
                          -offset=> $offset,
                          -count => $count,
                          -fill  => $fill,
+                         -filltag => $filltag,
 			 -total => \$total,
                          -long  => 1,
 			);
@@ -1272,6 +1597,68 @@ ACEDB requires.  date() will truncate the time portion.
 
 If not provided, $time defaults to localtime().
 
+=head1 OTHER METHODS
+
+=head2 debug()
+
+  $debug_level = Ace->debug([$new_level])
+
+This class method gets or sets the debug level.  Higher integers
+increase verbosity.  0 or undef turns off debug messages.
+
+=head2 name2db()
+
+ $db = Ace->name2db($name [,$database])
+
+This class method associates a database URL with an Ace database
+object. This is used internally by the Ace::Object class in order to
+discover what database they "belong" to.
+
+=head2 cache()
+
+Get or set the Cache::SizeAwareFileCache object, if one has been
+created.
+
+=head2 memory_cache_fetch()
+
+  $obj = $db->memory_cache_fetch($class,$name)
+
+Given an object class and name return a copy of the object from the
+in-memory cache.  The object will only be cached if a copy of the
+object already exists in memory space.  This is ordinarily called
+internally.
+
+=head2 memory_cache_store($obj)
+
+Store an object into the memory cache.  This is ordinarily called
+internally.
+
+=head2 memory_cache_delete($obj)
+
+Delete an object from the memory cache. This is ordinarily called
+internally.
+
+=head2 memory_cache_clear()
+
+Completely clears the memory cache.
+
+=head2 file_cache_fetch()
+
+  $obj = $db->file_cache_fetch($class,$name)
+
+Given an object class and name return a copy of the object from the
+file cache.  This is ordinarily called internally.
+
+=head2 file_cache_store($obj)
+
+Store an object into the file cache.  This is ordinarily called
+internally.
+
+=head2 file_cache_delete($obj)
+
+Delete an object from the file cache.  This is ordinarily called
+internally.
+
 =head1 THE LOW LEVEL C API
 
 Internally Ace.pm makes C-language calls to libace to send query
@@ -1336,7 +1723,7 @@ Returns a more detailed error code supplied by the Ace server.  Check
 this value when STATUS_ERROR has been returned.  These constants are
 also exported by default.  Possible values:
 
- ACE_INVALID           
+ ACE_INVALID
  ACE_OUTOFCONTEXT
  ACE_SYNTAXERROR
  ACE_UNRECOGNIZED
@@ -1394,6 +1781,13 @@ disclaimers of warranty.
 
 # -------------------- AUTOLOADED SUBS ------------------
 
+sub debug {
+  my $package = shift;
+  my $d = $DEBUG_LEVEL;
+  $DEBUG_LEVEL = shift if @_;
+  $d;
+}
+
 # Return true if the database is still connected.  This is oddly convoluted
 # because there are numerous things that can go wrong, including:
 #   1. server has gone away
@@ -1608,58 +2002,25 @@ sub find {
   $filled ? $self->_fetch($count,$offset) : $self->_list($count,$offset);
 }
 
-# Count the objects matching pattern without fetching them.
-sub count {
-  my $self = shift;
-  my ($class,$pattern,$query) = rearrange(['CLASS',
-					   ['NAME','PATTERN'],
-					   'QUERY'],@_);
-  $Ace::Error = '';
-
-  # A special case occurs when we have already fetched this
-  # object and it is already on the active list.  In this
-  # case, we do not need to recount.
-  $query   = '' unless defined $query;
-  $pattern = '' unless defined $pattern;
-  $class   = '' unless defined $class;
-
-  my $active_tag = "$class$pattern$query";
-  if (defined $self->{'active_list'} &&
-      defined ($self->{'active_list'}->{$active_tag})) {
-    return $self->{'active_list'}->{$active_tag};
-  }
-
-  if ($query) {
-    $query = "query $query" unless $query=~/^query\s/;
-  } else {
-    $pattern =~ tr/\n//d;
-    $pattern ||= '*';
-    $pattern = Ace->freeprotect($pattern);
-    $query = "find $class $pattern";
-  }
-  my $result = $self->raw_query($query);
-#  unless ($result =~ /Found (\d+) objects/m) {
-  unless ($result =~ /(\d+) Active Objects/m) {
-    $Ace::Error = 'Unexpected close during find';
-    return;
-  }
-  return $self->{'active_list'}->{$active_tag} = $1;
-}
-
 #########################################################
 # Grep function returns count in a scalar context, list
 # of retrieved objects in a list context.
 sub grep {
   my $self = shift;
-  my ($pattern,$count,$offset,$filled,$total,$long) = 
-      rearrange(['PATTERN','COUNT','OFFSET',['FILL','FILLED'],'TOTAL','LONG'],@_);
+  my ($pattern,$count,$offset,$filled,$filltag,$total,$long) = 
+      rearrange(['PATTERN','COUNT','OFFSET',['FILL','FILLED'],'FILLTAG','TOTAL','LONG'],@_);
   $offset += 0;
   my $grep = defined($long) && $long ? 'LongGrep' : 'grep';
   my $r = $self->raw_query("$grep $pattern");
   my ($cnt) = $r =~ /Found (\d+) objects/m;
   $$total = $cnt if defined $total;
   return $cnt if !wantarray;
-  return $filled ? $self->_fetch($count,$offset) : $self->_list($count,$offset);
+  if ($filltag) {
+    @h = $self->_fetch($count,$offset,$filltag);
+  } else {
+    @h = $filled ? $self->_fetch($count,$offset) : $self->_list($count,$offset);
+  }
+  @h;
 }
 
 sub pick {
